/******************************************************************************/
/*									      */
/* ls - a Unix-like directory listing program for MS-DOS 2.x		      */
/*									      */
/* Copyright (c) 1984  R. Edward Nather					      */
/* Copyright (c) 1985  Larry A. Shurr					      */
/*									      */
/******************************************************************************/
#include <stdio.h>

typedef int bool;			/*Define "boolean" type*/
typedef int mchar;			/*Define "metachar" type*/

char*	calloc();			/*Assure that calloc() is char* type*/
char*	malloc();			/*Assure that malloc() is char* type*/
char*	strrchr();			/*Declare the strrchr() function*/

#define FALSE		0		/*Define "false" value*/
#define TRUE		1		/*Define "true"  value*/

#define ATTR_ALL	0x1F		/*All attribute mask*/
#define ATTR_DIR	0x10		/*Directory attribute mask*/
#define ATTR_NUL	0x00		/*Null attribute mask*/
#define ATTR_VOL	0x08		/*Volume label attribute*/
/******************************************************************************/
			/* customizing constants */

#define NAMESIZ 13			/*12 character name + NULL*/
#define ONECS 512			/*cluster size on one-sided floppy*/
#define TWOCS 1024			/*cluster size on two-sided floppy*/
#define HARDCS 4096			/*cluster size on hard disk*/
#define SCRSIZ 22			/*scrolling size of display screen*/
/******************************************************************************/
struct dta {				/*DOS Disk Transfer Address table*/
  char	reserved[21];			/*used in "find next" operation*/
  char	attr;				/*file attribute byte*/
  int	ftime;				/*time of last modification*/
  int	fdate;				/*date of last modification*/
  long	fsize;				/*file size in bytes*/
  char	fname[NAMESIZ];			/*filename and extension*/
};

typedef struct dta DTA;
/******************************************************************************/
struct llst {				/*structure to hold file information*/
  char	*fattr;				/*file attribute pointer*/
  long	size;				/*file size*/
  int	day;				/*the day of creation*/
  int	mnum;				/*month*/
  int	yr;				/*year*/
  int	hh;				/*hour*/
  int	mm;				/*minute*/
  int	ap;				/*am or pm*/
};

typedef struct llst LLST;
/******************************************************************************/
struct obuf {				/*output buffer -- array of file info*/
  unsigned oattr;			/*file attributes*/
  unsigned odate;			/*file creation/modification date*/
  unsigned otime;			/*file creation/modification time*/
  long     osize;			/*file size in bytes*/
  char     oname[NAMESIZ+1];		/*file name*/
};

typedef struct obuf OBUF;

#define OBUFSIZ sizeof(obuf[0])
/******************************************************************************/
struct sysreg {				/*Generic register file*/
    int ax;
    int bx;
    int cx;
    int dx;
    int si;
    int di;
    int ds;
    int es; 
};

typedef struct sysreg SYSREG;
/******************************************************************************/
 struct sysregdx {			/*Register file with dx = char**/
  int   ax;
  int   bx;
  int   cx;
  char* dx;
  int   si;
  int   di;
  int   ds;
  int   es;
};

typedef struct sysregdx SYSREGDX;
/******************************************************************************/
mchar	qs       = '\\';		/*filename separator character*/
char	dqs[]    = "\\";		/*filename separator (string)*/

					/*"Dynamic" default flags (i.e.,*/
					/*defaults set in the environment)*/

bool	df_all   = FALSE;		/*1 => show hidden files by default*/
bool	df_colm  = FALSE;		/*1 => 1-column listing by default*/
bool	df_du    = FALSE;		/*1 => include disk use by default*/
bool	df_id    = FALSE;		/*1 => always identify directory*/
bool	df_long  = FALSE;		/*1 => long listing by default*/
bool	df_more  = FALSE;		/*1 => pause for more*/
bool	df_rsort = FALSE;		/*1 => reverse sort by default*/
bool	df_tsort = FALSE;		/*1 => time sort by default*/

					/*Opts intialized by dynamic defaults*/

int	f_all;				/*Include hidden & system files*/
int	f_colm;				/*1-column format*/
int	f_du;				/*Print disk usage*/
int	f_long;				/*Long listing*/
int	f_rsort;			/*Reverse sort*/
int	f_tsort;			/*Timesort the listing*/

					/*Other option flags*/

bool	f_help  = FALSE;		/*Help screen requested*/
bool	f_more  = FALSE;		/*Pause for more*/
bool	f_recd  = FALSE;		/*Recursive descent requested*/
bool	f_so    = FALSE;		/*Print sizes only*/
bool	f_tsc   = FALSE;		/*Output is to console screen*/
bool	f_to    = FALSE;		/*Print totals only for -s option*/

					/*Other program flags*/

bool	f_multi = FALSE;		/*Multiple patterns specified*/

long	left;				/*unused space left on disk*/
long	total;				/*total of sizes encountered*/

int	clsize	= 0;			/*size of a cluster, in bytes*/
int	clmask;				/*clsize-1 for rounding & chopping*/
int	drive   = -1;			/*code number for drive requested*/
int	lc      = 0;			/*line counter for endlin()*/
int	np;				/*number of groups printed*/
OBUF*	obuf;				/*Directory information buffer*/
int	obufk   = 1400;			/*Number of obuf entries*/
/******************************************************************************/
main(argc,argv)
int	argc;
char**	argv;
{
  int	res;				/*Gets function results if needed*/

  init();				/*Initialize*/

  res = getoptions(&argc,&argv);	/*Get options from command line*/

  if (res < 0 || f_help) {		/*If help requested...*/
    give_help();			/*Print the help stuff*/
    return;				/*And quit now*/
  }					/*Otherwise, continue...*/

  if (argc == 0) {			/*If no arguments...*/
    strcpy(*argv,"");			/*Set up a null argument*/
    argc += 1;				/*Say that there is one argument*/
  }					/*Otherwise, continue...*/

  while (argc-- > 0) {			/*While arguments remain...*/
    char* nxt = *argv++;		/*Get argument and advance arg pointer*/
    if (getsize(nxt) < 0) continue;	/*Get usage data, if drv bad, skip dir*/
    search(nxt);			/*Search and display directory*/
    if (argc == 0) {			/*If we've handled all arguments...*/
      if (f_du || f_so) {		/*If usage data was requested...*/
        if (np > 1) {			/*If more than directory listed...*/
          fprintf(			      /*Print a summation...*/
            stdout,			      /*To the standard output file*/
            "--------\n%8ld bytes total",     /*Message text*/
            total			      /*Total usage data*/
          );				      /*And continue...*/
          if(!f_to) fputs("; ",stdout);	/*If not just totals only, add "; "*/
        }				/*Otherwise, continue...*/
        if(!f_to) {			/*If not just totals only...*/
          fprintf(			      /*Print space remaining...*/
            stdout,			      /*To standard output file*/
            "%8ld bytes left on drive %c:\n", /*Message text*/
            left,			      /*Space remaining*/
            drive+'a'			      /*Drive name*/
          );				      /*And continue...*/
        }				/*Otherwise, continue...*/
      }					/*Otherwise, continue...*/
    }					/*Otherwise, continue...*/
  }
}
/******************************************************************************/
/* Name:	comp - COMPare two directory entries			      */
/*									      */
/* Synopsis:	result = comp(a,b)					      */
/*									      */
/*		OBUF*	a;		First directory entry to compare      */
/*		OBUF*	b;		Second directory entry to compare     */
/*		int	result;		< 0 => a < b			      */
/*					= 0 => a = b			      */
/*					> 0 => a > b			      */
/*									      */
/* Abstract:	Compare the two obuf directory entries according to the	      */
/*		comparison options set for this run: Default is alphanumeric  */
/*		comparison of the file names.  If the f_tsort flag is set     */
/*		the time/date's of the two files are compared and, by	      */
/*		default, the file with the earlier date is "less than" the    */
/*		other.  If the f_rsort flag is set, the sort direction is     */
/*		reversed.						      */
/******************************************************************************/
comp(a,b)				/*compare two filename entries*/
OBUF*	a;
OBUF*	b;
{
  int y;

  if (f_tsort) {			/*If sort-by-time selected...*/
    if (a->odate != b->odate)		  /*If dates differ...*/
      y = (a->odate < b->odate) ? -1 : 1; /*Compare dates*/
    else				  /*Otherwise...*/
      y = (a->otime < b->otime) ? -1 : 1; /*Compare times*/
    return((f_rsort) ? y : -y);		  /*Return result*/
  }					/*Otherwise...*/
  else {				/*sort-by-name is selected, so...*/
    y = strcmp(a->oname,b->oname);	/*Compare filenames*/
    return((f_rsort) ? -y : y);		/*Return result*/
  }					/*And continue...*/
}
/******************************************************************************/
/* Name:	endlin - produce END-of-LINe characters with pause on console */
/*									      */
/* Synopsis:	endlin()						      */
/*									      */
/* Abstract:	Called every time we wish to end a line of output to the      */
/*		the standard output file.  If the standard output file is     */
/*		the console and the f_more flag is set, we pause after 22     */
/*		lines have been output and prompt the user with "--More--".   */
/*									      */
/*		After the user responds to the prompt with a single key-      */
/*		stroke, we erase the prompt and leave the cursor positioned   */
/*		at the beginning of the current line of the display.	      */
/*									      */
/*		If the user responds to the prompt with a carriage return,    */
/*		we set up to allow only a single line of output before	      */
/*		pausing again, and return a blank instead of an end-of-line/. */
/*									      */
/*		If the user responds with a "q", we terminate execution of    */
/*		the program immediately.				      */
/*									      */
/*		If the user responds with anything else, we reset the line    */
/*		counter, permitting another 22 lines of output before we      */
/*		pause again and return a blank.				      */
/*									      */
/*		If standard output is not to the console, we do not pause     */
/*		and we always return an end-of-line instead.		      */
/******************************************************************************/
endlin()				/*End line & watch for screen full*/
{
  int	key;				/*Gets user key-in*/

  if (f_tsc  &&				/*If output is to screen and...*/
      f_more &&				/*If pause for more flag set and...*/
      ++lc >= SCRSIZ) {			/*If we have shown a screenful...*/
    fputs("\n--More--",stdout);		/*Pause and...*/
    fflush(stdout);			/*Force output to standard output*/
    key = ci();				/*Await user action code*/

    fputs("\b\b\b\b\b\b\b\b        \b\b\b\b\b\b\b",stdout); /*Blank it out*/

    switch(key) {			/*Do what user asked for...*/
      case '\r':			/*<RETURN> - show 1 more line*/
	lc = SCRSIZ - 1;
	break;
      case 'q':				/*quit with "q" or "ctrl-C"*/
      case '\003':
	exit(0);
      default:				/*else show another screenful*/
	lc = 0;
	break;

    }
    return('\b');
  }	
  else return('\n');
}
/******************************************************************************/
fill(i,ll)				/*fill long list structure*/
int i;
struct llst *ll;
{
  int j, k;
  static char fbuf[16][4] = {		/*Symbolic attribute definitions*/
    "--w", "---", "-hw", "-h-",
    "s-w", "s--", "shw", "sh-",
    "d-w", "d--", "dhw", "dh-",
    "d-w", "d--", "dhw", "dh-"
  };

  if ((obuf[i].oattr & 0x10) &&		/*If file is a directory and...*/
      obuf[i].oname[0] != '.') {	/*it is not a directory alias...*/
    ll->size = clsize;			/*Assume size is disk cluster size*/
    j = 8;				/*Set the dir symbolic attr offset*/
  }					/*Otherwise...*/
  else {				/*File is a regular file...*/
    ll->size = obuf[i].osize;		/*Set file size*/
    j = 0;				/*Set file symbolic attr offset*/
  }
  ll->fattr = fbuf[(obuf[i].oattr & 0x07) + j];	  /*Set symbolic attribute*/
  ll->day = obuf[i].odate & 0x1F;		  /*Set the day of the month*/
  ll->mnum = (obuf[i].odate >> 5) & 0x0F;	  /*Set the month*/
  ll->yr = (obuf[i].odate >> 9) + 1980;		  /*Set the year*/
  k = obuf[i].otime >> 5;			  /*Get hours and minutes*/
  ll->mm = k & 0x3f;				  /*Set minute*/
  ll->ap = ((ll->hh = k >> 6) >= 12) ? 'p' : 'a'; /*Set AM or PM*/
  if (ll->hh > 12) ll->hh -= 12;		  /*Convert to 12-hour time*/
  if (ll->hh == 0) ll->hh = 12;			  /*Convert midnite to 12 AM*/
}
/******************************************************************************/
/* Name:	fnd_1st - FiND 1ST match for file in directory		      */
/*									      */
/* Synopsis:	result = fnd_1st(path,dta,mask)				      */
/*									      */
/*		char*	path;		Pathname/filename pattern for search  */
/*		DTA*	dta;		Find file data structure	      */
/*		int	mask;		MSDOS file attribute mask for search  */
/*		bool	result;		TRUE if match found, FALSE otherwise  */
/*									      */
/* Abstract:	Find the first matching file for the specified pathname/file- */
/*		name pattern using the MSDOS find first matching file func-   */
/*		tion code.						      */
/******************************************************************************/
bool fnd_1st(path,dta,mask)		/*find first file in directory*/
char	*path;
DTA*	dta;
int	mask;
{
  SYSREGDX r;				/*Register file for FIND FIRST*/

  setdta(dta);				/*Set up disk transfer address*/

					/*Prepare for find first function...*/
  r.ax = 0x4E00;			/*Set find first function code*/
  r.cx = mask;				/*Attribute mask for search*/
  r.dx = path;				/*File pathname pattern*/
  r.ds = _showds();			/*Set up local DS*/

  return(!(sysint(0x21,&r,&r) & 1));	/*fills the structure*/
}
/******************************************************************************/
/* Name:	fnd_nxt - FiND NeXT match for file in directory		      */
/*									      */
/* Synopsis:	result = fnd_nxt1st(dta)				      */
/*									      */
/*		DTA*	dta;		Find file data structure initialized  */
/*					by a call to fnd_1st()		      */
/*		bool	result;		TRUE if match found, FALSE otherwise  */
/*									      */
/* Abstract:	Find the next matching file for the pathname/filename pattern */
/*		set up by a call to the fnd_1st function.  Each time a call   */
/*		to fnd_nxt() with an initialized dta is made, the next 	      */
/*		matching filename is returned until all matching files have   */
/*		been found.						      */
/******************************************************************************/
bool fnd_nxt(dta)			/*find next file in directory*/
DTA*	dta;
{
  SYSREG r;				/*Register file for FIND NEXT*/

  setdta(dta);				/*Set up disk transfer address*/

					/*Prepare for find next function...*/
  r.ax = 0x4F00;			/*Set 'find next' function code*/

  return(!(sysint(0x21,&r,&r) & 1));	/*fill the table*/
}
/******************************************************************************/
/* Name:	getcdate - GET Current DATE				      */
/*									      */
/* Synopsis:	result = getcdate()					      */
/*									      */
/*		int	result;		current date in months		      */
/*									      */
/* Abstract:	Return current date in months (i.e., current year * 12 +      */
/*		current month).						      */
/******************************************************************************/
getcdate()
{
  SYSREG r;				/*Register file for GET DATE*/

  r.ax = 0x2A00;			/*Set 'get date' function code*/
  sysint(0x21,&r,&r);			/*Call DOS*/
  return(r.cx * 12 + (r.dx >> 8));	/*Return date in months*/
}
/******************************************************************************/
/* Name:	getcdrv - GET Current DRiVe				      */
/*									      */
/* Synopsis:	getcdrv(sp)						      */
/*									      */
/*		char*	sp;		String to receive the drive name      */
/*									      */
/* Abstract:	Obtain the currently selected drive number and construct      */
/*		a drive name string in the provided string buffer in the      */
/*		format:							      */
/*									      */
/*			"<drive>:"					      */
/******************************************************************************/
getcdrv(sp)
char*	sp;
{
  SYSREG r;				/*Register file for GET DRV*/

  r.ax = 0x1900;			/*Set 'get current drive' code*/
  sysint(0x21,&r,&r);			/*Call DOS*/
  *sp++ = r.ax + 'a';			/*Convert to symbolic drive name*/
  *sp++ = ':';				
  *sp   = 0;
}
/******************************************************************************/
/* Name:	getcl - GET disk CLuster size and space remaining on disk     */
/*									      */
/* Synopsis:	size = getcl(pp)					      */
/*									      */
/*		char*	pp;		Pathname string			      */
/*		int	size;		Cluster size for specified disk	      */
/*									      */
/* Abstract:	The cluster size is obtained for the disk drive specified     */
/*		in the pathname.  If no drive name is specified in the	      */
/*		pathname, the default drive is assumed.			      */
/*									      */
/* Side affects:The global mchar, drive, is set to the drive number of the    */
/*		accessed disk drive.					      */
/*									      */
/*		The global long integer, left, is set to the amount of space, */
/*		in bytes, available on the disk drive.			      */
/*									      */
/* Notes:	If the requested drive is the same as the last drive, we      */
/*		return the last cluster size obtained and we do not recompute */
/*		the space available on the last drive either.		      */
/******************************************************************************/
int getcl(pp)
char *pp;
{
  int cs;

  SYSREG r;				/*Register file for GET FREE SPACE*/

  if (*(pp+1) == ':') {			/*If a drive was specified...*/
    r.ax = *pp - 'a';			/*Use specified drive id*/
  }					/*Otherwise...*/
  else {				/*Obtain current drive...*/
    r.ax = 0x1900;			/*Set 'get default drive' code*/
    sysint(0x21,&r,&r);			/*Call MSDOS*/
    r.ax = r.ax & 0x7F;			/*Trim extraneous data*/
  }					/*And continue...*/

  if (drive != -1) {			/*If no previous drive setting...*/
    if (drive != r.ax) {		/*If drive is different from last...*/
      f_to = TRUE;			/*Force totals only flag on*/
    }					/*Otherwise drive is same as last...*/
    else return(clsize);		/*So return the previous cluster size*/
  }					/*Otherwise, continue...*/

  drive = r.ax;				/*Save the search drive code*/

  r.ax = 0x3600;			/*Set 'get disk free space' function*/
  r.dx = drive + 1;			/*Set drive for 'get disk free space'*/
  sysint(0x21,&r,&r);			/*Get disk free space and cluster size*/
  if (r.ax == 0xFFFF) return(0);	/*If drive is invalid, error return*/
  cs = r.ax * r.cx;			/*Compute cluster size in bytes*/
  left = (long)cs * r.bx;		/*Compute disk free space in bytes*/
  return(cs);				/*Return the cluster size*/
}
/******************************************************************************/
/* Name:	getdir - GET current DIRectory				      */
/*									      */
/* Synopsis:	result = getdir(ext,pro)				      */
/*									      */
/*		char*	ext;		Extended directory pathname (output)  */
/*		char*	pro;		Prototype directory pathname (input)  */
/*		int	result		< 0 if an error occurs		      */
/*									      */
/* Abstract:	Expand the prototype directory pathname to it's full (or      */
/*		extended) name.  If the prototype has a "tail" consisting     */
/*		of the pathname separator ("/" or "\") and/or a filename      */
/*		pattern, it is appended to the extended pathname. 	      */
/******************************************************************************/
getdir(ext,pro)
char*	ext;
char*	pro;
{
  DTA	dta;				/*Dummy DTA structure for dir search*/
  bool	f_dir_exists = FALSE;		/*"Directory exists flag*/
  char	lpro[80];			/*Local copy of prototype*/
  int	plen         = 0;		/*Length of prototype string*/
  char	save[80];			/*For saving drive's current directory*/
  char	tailb[20];			/*Retains a copy of the path tail*/
  char*	tailp;				/*For locating the end of the dir path*/

  ext[0] = 0;				/*Null the extended directory string*/
  tailb[0] = 0;				/*Null the tail buffer*/

	    /*Determine if prototype specifies a file or directory*/

  strcpy(lpro,pro);			/*Make local copy of prototype string*/
  plen = strlen(lpro);			/*Compute length of prototype*/
  if (plen > 0) {			     /*If prototype is non-null...*/
    if (lpro[plen-1] == qs) {		     /*If it ends with "\"...*/
      if (plen > 1 && lpro[plen-2] != ':') { /*If it is not root dir...*/
        lpro[plen-1] = 0;		     /*Delete trailing "\"*/
        if (fnd_1st(lpro,&dta,ATTR_DIR)) {   /*If the file exists...*/
          if (dta.attr & ATTR_DIR) {	     /*And it is a directory...*/
            f_dir_exists = TRUE;	     /*Set "dir exists" flag*/
            strcat(lpro,dqs);		     /*Append a new "\"*/
          }				     /*Otherwise, path not dir...*/
          else return(-1);		     /*So error return*/
        }				     /*Otherwise, path not dir...*/
        else return(-1);		     /*So error return*/
      }					     /*Otherwise, it is root dir...*/
      else f_dir_exists = TRUE;		     /*So set "dir exists" flag*/
    }					     /*Otherwise...*/
    else {				     /*Path doesn't end with "\"...*/
      char* w = strrchr(lpro,'*');	     /*Find any wildcards*/
      if (w == NULL) w = strrchr(lpro,'?');  /*If no wildcard, try again*/      
      if (w == NULL) {			     /*If no wildcard...*/
        if (fnd_1st(lpro,&dta,ATTR_DIR)) {   /*If the file exists...*/
          if (dta.attr & ATTR_DIR) {	     /*And it is a directory...*/
            f_dir_exists = TRUE;	     /*Set "dir exists" flag*/
            strcat(lpro,dqs);		     /*Append a "\"*/
          }				     /*Otherwise, continue...*/
        }				     /*Otherwise, continue...*/
      }					     /*Otherwise, continue...*/
    }					     /*And continue...*/
  }					/*Otherwise...*/
  plen = strlen(lpro);			/*Recompute length of prototype*/

		 /*Copy the prototype to the extended string,*/
		      /*adding the drive name if needed*/

  if (plen > 1) {			/*If prototype long enough for dvc...*/
    if (lpro[1] != ':') getcdrv(ext);	/*If no device spec, set current drive*/
  }					/*Otherwise, it's not long enough...*/
  else getcdrv(ext);			/*So set current drive*/
  strcat(ext,lpro);			/*Copy prototype path to extended*/

		  /*Locate the tail of the pathname, save it*/
		   /*and delete it from the directory path*/

  tailp = strrchr(ext,qs);		/*Locate the end of the dir path*/
  if (tailp != NULL) {			/*If found...*/
    if (*(tailp-1) == ':') {		/*If directory is the root,...*/
      tailp += 1;			/*Include "\" in path*/
      f_dir_exists = TRUE;		/*Assume that directory exists*/
    }					/*Otherwise, continue...*/
  }					/*Otherwise...*/
  else {				/*No directory in path, (=> current)*/
    f_dir_exists = TRUE;		/*Assume current directory exists*/
    tailp = ext + 2;			/*Tail begins after the device name*/
  }					/*And continue...*/
  strcpy(tailb,tailp);			/*Copy the tail to a save buffer*/
  *tailp = 0;				/*Trim non-directory data from string*/

		/*Determine if the specified directory exists*/

  if (!f_dir_exists) {			/*If we don't know yet...*/
    if (fnd_1st(ext,&dta,ATTR_DIR)) {	/*If the file exists...*/
      if (dta.attr & ATTR_DIR) {	/*And it is a directory...*/
        f_dir_exists = TRUE;		/*Set "directory exists" flag*/
      }				     	/*Otherwise, continue...*/
    }					/*Otherwise, continue...*/
  }					/*Otherwise, continue...*/

		     /*If we haven't found the directory*/
		        /*we return an error condition*/

  if (!f_dir_exists) {			/*If we haven't found directory...*/
    ext[0] = 0;				/*Null extended directory string*/
    return(-1);				/*Error return*/
  }					/*Otherwise, continue...*/

	       /*Get the drive's current directory and save it*/

  strcpy(save,ext);			/*Copy temporary ext string to save*/
  getpath(save);			/*Obtain and save drive's current dir*/

		  /*Extend (expand) the specified directory*/

  setpath(ext);				/*Set drive's default dir to that*/
  getpath(ext);				/*Obtain the fully extended dir name*/

		/*Restore the drive's "real" current directory*/

  setpath(save);			/*Restore the drive's current dir*/

	      /*Edit the extended directory pathname for beauty*/

  if (tailb[0] != qs) {			/*If tail doesn't being with "\"...*/
    if (ext[strlen(ext)-1] != qs) {	/*If ext name doesn't end with "\"...*/
      strcat(ext,dqs);			/*Append a "\" to the extended name*/
    }					/*Otherwise, continue...*/
  }					/*Otherwise, continue...*/
  if (ext[strlen(ext)-1] == qs) {	/*If the extended name is "<dev>:\"...*/
    if (tailb[0] == qs) {		/*And if the tail begins with "\"...*/
      strcpy(tailb,(char*)&tailb[1]);	/*Delete the leading "\" in the tail*/
    }					/*Otherwise, continue...*/
  }					/*Otherwise, continue...*/
  strcat(ext,tailb);			/*Append the tail to the extended name*/
  while (*ext) {			/*Final scan of extended string...*/
    *ext = tolower(*ext);		/*Convert each character to lower case*/
    if (*ext == '/' || *ext == '\\') {	/*If char is a pathname separator...*/
      *ext = qs;			/*Force to be the user designated char*/
    }					/*Otherwise, continue...*/
    ext += 1;				/*Advance to the next character*/
  }					/*End of final scan*/
  return(0);				/*Return a "no error" condition*/
}
/******************************************************************************/
/* Name:	getoptions - GET OPTIONS from command line		      */
/*									      */
/* Synopsis:	getoptions(argc,argv)					      */
/*									      */
/*		int*	argc;		Number of command line arguments      */
/*		char**	argv[];		Argument vector			      */
/*									      */
/* Abstract:	Scan the command line arguments (if any) and set option	      */
/*		flags accordingly.					      */
/*									      */
/* Side affects:The argument count and argument vector pointers (passed as    */
/*		parameters) are modified to "eliminate" the option flags      */
/*		from the command line.					      */
/*									      */
/*		Sets the global flag, f_multi if the number of non-option     */
/*		arguments (i.e., the number of file pathname patterns)	      */
/*		specified on the command line is greater than one.	      */
/******************************************************************************/
getoptions(argc,argv)
int*	argc;
char**	argv[];
{
  int	defective = 0;			/*Defective options counter*/
  char	*s;				/*Option string scanning pointer*/

  while((--*argc > 0) &&		/*While we keep finding arguments...*/
        (*++*argv)[0] == '-') {		/*which are option flags...*/
    for(s = *argv[0]+1; *s; s++) {	/*Scan the option string...*/
      switch(*s) {			/*Select according to option...*/
        case 'a':			/*-a: list all files*/
          f_all   = !df_all;		/*Invert list all files default*/
          break;				
        case 'c':			/*-c: 1-column listing requested*/
          f_colm  = !df_colm;		/*Invert 1-column listing default*/
          break;
        case 'h':			/*-h: print help and quit*/
          f_help  = TRUE;		/*Set the help flag*/
          break;
        case 'l':			/*-l: long listing requested*/
          f_long  = !df_long;		/*Invert the long listing default*/
          break;
        case 'm':			/*-m: pause for more*/
          f_more  = !df_more;		/*Invert the pause for more default*/
          break;
        case 'r':			/*-r: reverse sort direction*/
          f_rsort = !df_rsort;		/*Invert the reverse sort default*/
          break;
        case 's':			/*-s: print sizes only*/
          f_so = TRUE;			/*Set the print sizes only flag*/
          if (*(s+1) == '1') {		/*If a '1' is appended...*/
            clsize = ONECS;		/*Set cluster size for 1-sided floppy*/
            f_to = TRUE;		/*Set the total only flag*/
            s++;			/*Advance the option string pointer*/
          }				/*Otherwise...*/
          else if (*(s+1) == '2') {	/*If a '2' is appended...*/
            clsize = TWOCS;		/*Set cluster size for 2-sided floppy*/
            f_to = TRUE;		/*Set the total only flag*/
            s++;			/*Advance the option string pointer*/
          }				/*Otherwise, continue...*/
          break;
        case 't':			/*-t: time sort requested*/
          f_tsort = !df_tsort;		/*Invert the time sort default*/
          break;
        case 'u':  			/*-u: print disk usage*/
          f_du    = !df_du;		/*Invert the print disk usage default*/
          break;
        case 'R':			/*-R: recursively list subdirs*/
          f_recd  = TRUE;		/*Set the recursive list flag*/
          break;
        default:
          fprintf(stderr,"unknown arg %c\n",*s);
          defective += 1;		
      }
    }
  }

  f_multi = *argc > 1;			/*Setup multiple dir arguments flag*/

  return(defective == 0 ? 0 : -1);	/*Return -1 if error in any option*/
}
/******************************************************************************/
/* Name:	getpath - GET default PATHname for drive		      */
/*									      */
/* Synopsis:	getpath(sp)						      */
/*									      */
/*		char*	sp;		String to receive pathname	      */
/*					Must be initialized with drive name   */
/*									      */
/* Abstract:	Obtains default pathname for the specified drive.	      */
/*									      */
/* Notes:	The parameter sp must be initialized with the drive name      */
/*		thusly: "<drive>:" where <drive> ::= a valid drive letter.    */
/*									      */
/*		The sp parameter may contain additional data following the    */
/*		drive name without any problems, any such data is ignored.    */
/******************************************************************************/
getpath(sp)
char	*sp;
{
  struct {				/*Register file for GET PATH*/
    int   ax;
    int   bx;
    int   cx;
    int   dx;
    char* si;
    int   di;
    int   ds;
    int   es;
  } r;

  sp[2] = qs;				/*Put root file symbol after drive*/
  sp[3] = 0;				/*Terminate drive name string*/

					/*Prepare to get current path...*/
  r.ax = 0x4700;			/*Set 'get current directory' code*/
  r.dx = *sp - '`';			/*Set drive number*/
  r.ds = _showds();			/*Set pathname buffer segment*/
  r.si = sp + 3;			/*Set pathname buffer offset*/
  sysint(0x21,&r,&r);			/*Call DOS*/
}
/******************************************************************************/
getsize(path)				/*get file cluster size*/
char	*path;
{
  if ((clsize = getcl(path)) == 0) {	/*Get cluster size, if drv invalid...*/
    fprintf(				/*Print error message...*/
      stderr,				/*To the standard error file*/
      "Invalid drive: %c:\n",		/*Message text*/
      *path				/*Name of invalid drive*/
    );					/*And continue...*/
    np += 1;				/*Increment number of groups printed*/
    return(-1);				/*also, return error flag*/
  }					/*Otherwise, continue...*/
  clmask = clsize-1;			/*Set the cluster size mask as well*/
  return(0);				/*And return a 'no error' flag*/
}
/******************************************************************************/
/* Name:	give_help						      */
/*									      */
/* Synopsis:	give_help()						      */
/*									      */
/* Abstract:	Print a help summary and display the default settings.	      */
/******************************************************************************/
#define onoff(arg)  arg ? "ON " : "OFF"
#define forrev(arg) arg ? "REVERSE" : "FORWARD"

give_help()
{
  printf("\n");
  printf("      -a    Reverse the \"list all files\" option default\n");
  printf("      -c    Reverse the \"one-column listing\" option default\n");
  printf("      -h    Print this help message and the default settings\n");
  printf("      -l    Reverse the \"long listing\" option default\n");
  printf("      -m    Reverse the \"pause for more\" option default\n");
  printf("      -r    Reverse the default sort direction\n");
  printf("      -R    Recursively list sub-directories\n");
  printf("      -s    Print sizes\n");
  printf("      -s1   Print sizes for single-sided diskette\n");
  printf("      -s2   Print sizes for double-sided diskette\n");
  printf("      -t    Reverse the \"time sort\" option default\n");
  printf("      -u    Reverse the \"include disk usage\" option default\n");
  printf("\n");

  printf("List all files default is     %3s      ",onoff(df_all));
  printf("Default sort direction is     %7s  \n",forrev(df_rsort));

  printf("One-column listing default is %3s      ",onoff(df_colm));
  printf("Include disk usage default is %3s      \n",onoff(df_du));

  printf("Long listing default is       %3s      ",onoff(df_long));
  printf("Time sort option default is   %3s      \n",onoff(df_tsort));

  printf("Pause for more default is     %3s      ",onoff(df_more));
  printf("Always print dir default is   %3s      \n",onoff(df_id));

  printf("\n");
}
/******************************************************************************/
/* Name:	init - INITialize					      */
/*									      */
/* Synopsis:	init()							      */
/*									      */
/* Abstract:	Initialize global storage for the ls program.		      */
/******************************************************************************/
init()
{
  setbuf(stdout,malloc(BUFSIZ));	/*Force standard output buffered*/

  df_all   = envbool("LSALL",FALSE);	/*1 => show hidden files by default*/
  df_colm  = envbool("LSCOLM",FALSE);	/*1 => 1-column listing by default*/
  df_du    = envbool("LSDU",FALSE);	/*1 => include disk use by default*/
  df_id    = envbool("LSID",FALSE);	/*1 => always identify directory*/
  df_long  = envbool("LSLONG",FALSE);	/*1 => long listing by default*/
  df_more  = envbool("LSMORE",FALSE);	/*1 => pause for more*/
  df_rsort = envbool("LSRSORT",FALSE);	/*1 => reverse sort by default*/
  df_tsort = envbool("LSTSORT",FALSE);	/*1 => time sort by default*/

  qs = envchar("LSQS",qs);		/*Get default directory separator*/
  dqs[0] = qs;

  f_all	  = df_all;			/*include hidden & system files*/
  f_colm  = df_colm;			/*1-column format*/
  f_du	  = df_du;			/*print disk usage*/
  f_long  = df_long;			/*long listing*/
  f_more  = df_more;			/*pause for more*/
  f_rsort = df_rsort;			/*reverse sort*/
  f_tsort = df_tsort;			/*timesort the listing*/

  f_recd  = FALSE;			/*recursive descent requested*/
  f_so    = FALSE;			/*print sizes only*/

  f_tsc = toscreen();  			/*find out if output is to console*/

  do {
    obuf = (OBUF*)calloc(obufk,OBUFSIZ);/*Allocate the obuf*/
    if (obuf != NULL) break;		/*If we got it, quit trying*/
    obufk -= 100;			/*Otherwise, reduce request and...*/
  } while (obufk > 0);			/*Keep trying*/
}
/******************************************************************************/
LLST	l;

longlist(k)				/*list everything about the files*/
int	k;				/*total number to list*/
{

  int i, m, n, cdate;
  char *mon, *mname();

  cdate = getcdate();			/*get current date (in months)*/
  if (f_colm) {				
    n = k;				/*set for 1 column listing*/
  }
  else n = (k + 1)/2;			/*or for 2 column listing*/
  for (i = 0; i < n; i++) {
    for (m = 0; (m+i) < k; m += n) {		
      fill(i+m,&l);			/*fill list structure*/
      mon = mname(l.mnum);		/*conv month # to name*/
      fprintf(
        stdout,
        "%s%7ld  %2d %s ",
	l.fattr,
	l.size,
	l.day,
	mon
      );
      if (cdate >= (l.yr * 12 +l.mnum) + 12) {
	fprintf(stdout," %4d  ",l.yr);		/*print year if too old*/
      }
      else {
	fprintf(				/*else print time*/
          stdout,
          "%2d:%02d%c ",
	  l.hh,
          l.mm,
          l.ap
        );
      }
      putname(i+m);
      if (m+n < k) {
        fputs("\272 ",stdout);		/*double bar separator*/
      }
    }
    fputc(endlin(),stdout);
    if (f_tsc) fflush(stdout);
  }
}
/******************************************************************************/
/* Name:	mname - produce Month NAME				      */
/*									      */
/* Synopsis:	result = mname(n)					      */
/*									      */
/*		int	n;		Month number			      */
/*		char*	result;		Month name string		      */
/*									      */
/* Abstract:	Convert month number to month name			      */
/******************************************************************************/
char* mname(n)
int n;
{
  static char *name[] = {		/*Array of month names*/
    "???", "Jan", "Feb", "Mar",
    "Apr", "May", "Jun", "Jul",
    "Aug", "Sep", "Oct", "Nov",
    "Dec"
  };

  return((n < 1 || n > 12) ? name[0] : name[n]);
}
/******************************************************************************/
putname(i)				/*convert name to lower case and print*/
int i;
{
  int c, j = 0;

  while ((c = tolower(obuf[i].oname[j])) != 0) {
    fputc(c,stdout);
    j++;
  }
  while (j++ < NAMESIZ - 1) {		/*pad to columnarize*/
    fputc(' ',stdout);
  }
}
/******************************************************************************/
search(path)				/*search for filename or directory*/
char *path;
{
  long	bytes	   = 0;			/*Count of disk usage this directory*/
  DTA	dta;				/*DOS file data table*/
  bool	found      = FALSE;		/*"Found matching files" flag*/
  int	ndx;				/*Index for subdirectory list*/
  int	path_len;			/*Length of initial path*/
  int	k	   = 0;			/*Counts number of entries found*/
  int	mask	   = ATTR_DIR;		/*Attribute search mask*/
  char	tailb[20];			/*Tail buffer for recursive search*/
  char*	tailp;				/*Tail pointer for recursive search*/
  char	work[80];			/*Working path string*/

  static bool	f_negroup = FALSE;	/*Last group was non-empty flag*/

  int	comp();				/*Declare the compare function*/

  if (f_all) {				/*If find-all-files selected...*/
    mask = ATTR_ALL;			/*Set attr search mask appropriately*/
  }					/*Otherwise, continue...*/

  if (getdir(work,path) < 0) {		/*Get fully extended pathname*/
					/*If directory was not found...*/
    fprintf(					     /*Print error message...*/
      stderr,					     /*To standard error file*/
      "Directory is missing or invalid in \"%s\"\n", /*Tell what is wrong*/
      path				             /*Erroneous path spec*/
    );						     /*And continue...*/
    return;				/*Return immediately*/
  }					/*Otherwise, we have the directory*/

  path_len = strlen(work);		/*Get length of pathname pattern*/

  if (work[path_len-1] == qs) {		/*If no filename pattern specified...*/
    strcat(work,"*.*");			/*Append universal search pattern*/
  }					/*Otherwise, continue...*/

  if (fnd_1st(work,&dta,mask)) {	/*If a matching file was found...*/
    found = TRUE;			/*Set "matching files found" flag*/
    do {				/*Find & save all matches...*/
      if (dta.attr & ATTR_VOL) continue;/*Ignore volume label*/
      if (dta.fname[0] == '.' &&	/*If filename is "." or ".." and...*/
          !f_all) {			/*the -a option was not specified...*/
        continue;			/*Ignore these "files"*/
      }					/*Otherwise, continue...*/
      obuf[k].oattr = dta.attr;		/*Save file attributes*/
      obuf[k].otime = dta.ftime;	/*Save file time*/
      obuf[k].odate = dta.fdate;	/*Save file date*/
      obuf[k].osize = dta.fsize;	/*Save file size*/
      strcpy(obuf[k].oname,dta.fname);	/*Save file name*/
      if (f_du || f_so) {		/*If disk usage information needed...*/
        if((dta.attr & ATTR_DIR) &&	/*If the file is a directory and...*/
           dta.fname[0] != '.') {	/*It is not an alias name*/
          bytes += clsize;		/*Assume dir uses one cluster*/
        }				/*Otherwise...*/
        else if (dta.fsize) {		/*If file is non-zero length...*/
          obuf[k].osize =		/*Set file size to...*/
            ((dta.fsize + clmask) &	/*Number of clusters * cluster size...*/
             (long)(~clmask));		/*Used by the file*/
          bytes += obuf[k].osize;	/*And sum this usage data*/
        }				/*Otherwise, continue...*/
      }					/*Otherwise, continue...*/
      k++;				/*Increment files counter*/
    } while(fnd_nxt(&dta));		/*End of search and save loop*/
  }					/*Otherwise, continue...*/

  work[path_len] = NULL;		/*restore directory pathname*/

  if (found) {				/*If files were found...*/
    if (np++ && !f_so) {		/*If multi groups, but no size only...*/
      fputc(endlin(),stdout);		/*Separate listing blocks*/
    }					/*Otherwise, continue...*/
    if (f_du || f_so) {			/*If usage information requested...*/
      total += bytes;			/*Sum up total bytes used*/
      fprintf(stdout,"%8ld  ",bytes);	/*Print usage in present directory*/
    }					/*Otherwise, continue...*/
    if (f_recd  ||			/*If recursive descent requested...*/
        f_multi ||			/*Or multiple directories specified...*/
        f_du    ||			/*Or disk usage requested...*/
        f_so    ||			/*Or size only requested...*/
        df_id)  {			/*Or dir identification requested...*/
      fprintf(stdout,"%s",work);	/*Print directory name*/
      fputc(endlin(),stdout);		/*Terminate directory name line*/
      if (f_tsc) fflush(stdout);	/*Force line out to console*/
    }					/*Otherwise, continue...*/
    if (!f_so) {			  /*If directory info requested...*/
      hsort(obuf,k,sizeof(obuf[0]),comp); /*Sort the entries*/
      if (f_long)			  /*If long listing requested...*/
        longlist(k);			  /*Print directory in long format*/
      else				  /*Otherwise...*/
        shortlist(k);			  /*Print directory in short format*/
    }					  /*Otherwise, continue...*/
    f_negroup = TRUE;			/*Set last group non-empty flag*/
  }
  else {
    if (f_tsc) {			/*If output is to the console...*/
      np += 1;				/*Increment number of groups printed*/
      if (f_so) {			   /*If "size only" selected...*/
        fprintf(stdout," %7d  %s",0,work); /*Print zero usage and pathname*/
      }					   /*Otherwise...*/
      else {				/*We're going to print an error msg...*/
        if (f_negroup) {		/*If last group was non-empty...*/
          fputc(endlin(),stdout);	/*Output a blank line*/
        }				/*Otherwise, continue...*/
        fprintf(			/*Print no matching file error message*/
          stdout,			/*To standard output file*/
          "No match found for \"%s\"",	/*No files found message*/
          work				/*Tell pathname pattern*/
        );				/*And continue...*/
      }					/*And continue...*/
      fputc(endlin(),stdout);		/*End the message line*/
      if (f_tsc) fflush(stdout);	/*Force line out to console*/
    }					/*Otherwise...*/
    else fprintf(			/*Print error message*/
      stderr,				/*To standard error file*/
      "No match found for \"%s\"\n",	/*No files found message*/
      work				/*Tell pathname pattern*/
    );					/*And continue...*/
    f_negroup = FALSE;			/*Reset last group non-empty flag*/
  }					/*And continue...*/

  if (!f_recd) return;			/*If no recursive descent, return*/

  tailp = strrchr(work,qs);		/*Locate tail of directory pathname*/
  strcpy(tailb,tailp);			/*Save tail in the tail buffer*/
  tailp += 1;				/*Leaving the trailing "\"...*/
  *tailp = 0;				/*Trim tail from directory pathname*/
  strcat(work,"*.*");			/*Concatenate universal search pattern*/
  if (fnd_1st(work,&dta,mask)) {	/*And search for subdirectories...*/
    k = 0;				/*Initialize subdirectory counter*/
    do	{				/*Scan for subdirectories...*/
      if (dta.attr & ATTR_DIR &&	/*If we find a subdirectory...*/
          dta.fname[0] != '.') { 	/*Which is not an alias...*/
        strcpy(obuf[k].oname,dta.fname);/*Save subdirectory name...*/
        k += 1;				/*Increment subdirectories counter*/
      }					/*Otherwise, continue...*/
    } while(fnd_nxt(&dta));		/*End of scan for subdirectories...*/
    if (k > 0) {			/*If any subdirectories found...*/
      hsort((char*)obuf,k,OBUFSIZ,comp);/*Sort the subdirectories*/
      obuf += k;			/*Reserve subdirectory list in obuf*/
      for (ndx = 0; ndx < k; ndx++) {	/*Scan the subdirectory list...*/
        *tailp = 0;			/*Trim old tail from directory path*/
        strcat(work,obuf[ndx-k].oname);	/*Append subdirectory name to path*/
        strcat(work,tailb);		/*Append pathname tail after that*/
        search(work);			/*Search subdirectory*/
      }					/*Otherwise, continue...*/
      obuf -= k;			/*Unreserve subdirectory list in obuf*/
    }					/*End of subdirectory scan*/
  }					/*Otherwise no subdirectories...*/
}
/******************************************************************************/
/* Name:	setdta - SET Disk Transfer Address			      */
/*									      */
/* Synopsis:	setdta(dta)						      */
/*									      */
/*		DTA*	dta;		Address of MSDOS find file structure  */
/*									      */
/* Abstract:	Set the DTA to point to a find file structure for the	      */
/*		directory search.					      */
/******************************************************************************/
setdta(dta)
DTA*	dta;
{
  SYSREGDX r;				/*Register file for GET DTA*/

  r.ax = 0x1A00;			/*Set 'set disk transfer addr' func*/
  r.dx = (char*)dta;			/*Supply DTA offset*/
  r.ds = _showds();			/*Supply DTA segment*/
  sysint(0x21,&r,&r);			/*Call DOS*/
}
/******************************************************************************/
/* Name:	setpath - SET default PATHname for drive		      */
/*									      */
/* Synopsis:	setpath(sp)						      */
/*									      */
/*		char*	sp;		String to containing new pathname     */
/*					Must be initialized with drive name   */
/*									      */
/* Abstract:	Obtains default pathname for the specified drive.	      */
/*									      */
/* Notes:	If the pathname contains a drive specification, the default   */
/*		directory for the specified drive is set.		      */
/******************************************************************************/
setpath(sp)
char	*sp;
{
  SYSREGDX r;				/*Register file for SET PATH*/

					/*Prepare to set new directory path...*/
  r.ax = 0x3B00;			/*Set 'set new directory path' code*/
  r.dx = sp;				/*Set pathname buffer offset*/
  r.ds = _showds();			/*Set pathname buffer segment*/
  sysint(0x21,&r,&r);			/*Call DOS*/
}
/******************************************************************************/
/* Name:	shortlist - produce a SHORT (summary) LISTing of the files    */
/*									      */
/* Synopsis:	shortlist(k)						      */
/*									      */
/*		int	k;		Number of files recorded in *obuf     */
/*									      */
/* Abstract:	Produce a summary listing of the files in 1to 5 columns.      */
/******************************************************************************/
shortlist(k)
int	k;
{
  int i, m, n;

  if (f_colm) {				/*If single-column output selected...*/
    n = k;				/*Set for 1-column listing*/
  }					/*Otherwise...*/
  else n = (k + 4)/5;			/*Set for 2-to-5-column listing*/

  for (i = 0; i < n; i++) {		/*Scan the directory buffer...*/
    for (m = 0; (i+m) < k; m += n) {	/*Scan a row of data...*/
      if (obuf[i+m].oattr & 0x10) {	/*If file is a directory file...*/
	strcat(obuf[i+m].oname,dqs);	/*Mark it so*/
      }					/*Otherwise, continue...*/
      putname(i+m);			/*Print the name*/
      fputs("   ",stdout);		/*Output blanks to the next column*/
    }					/*End of scan-a-row*/
    fputc(endlin(),stdout);		/*Output an end-of-line to stdout*/
    if (f_tsc) fflush(stdout);		/*Force line out if stdout is console*/
  }					/*End of scan-directory-buffer*/
}
/******************************************************************************/
/* Name:	toscreen - discover is standard output is TO the SCREEN	      */
/*									      */
/* Synopsis:	result = toscreen()					      */
/*									      */
/*		bool	result;		TRUE if output is to screen	      */
/*					FALSE otherwise			      */
/*									      */
/* Abstract:	Use the I/O control system function to get the device	      */
/*		characteristics for standard output (handle 1).		      */
/*									      */
/******************************************************************************/
toscreen()				/*find out if output is to console*/
{
  SYSREG r;				/*Register file for IOCTL*/

  r.ax = 0x4400;			/*Set 'i/o control - get dvc info'*/
  r.bx = 1;				/*Set standard output handle*/
  sysint(0x21,&r,&r);			/*Call DOS*/
  return((int)r.dx & 1);		/*Return 'is console input' status*/
}
/******************************************************************************/
/*									      */
/*			 Compiler-specific code for			      */
/*									      */
/*			 Computer Innovations CI-86			      */
/*									      */
/******************************************************************************/
#ifdef CIC86


_showds()				/*Obtain local DS value*/
{
  struct {int cs, ss, ds, es;} r;

  segread(&r);
  return(r.ds);
}


ci()					/*Quick-and-dirty console input*/
{
  return(bdos(7) & 0xFF);
}


bool envbool(env,val)			/*Get environment boolean*/
char*	env;
bool	val;
{
  return(val);
}

mchar envchar(env,val)			/*Get environment character*/
char*	env;
mchar	val;
{
  return(val);
}

#else
/******************************************************************************/
/*									      */
/*			 Compiler-specific code for			      */
/*									      */
/*		       Lattice C version 2.0 and later			      */
/*									      */
/******************************************************************************/
#ifdef LATTICE

#include <dos.h>			/*DOS definitons*/

int	_stack = 3072;			/*Set run-time stack size*/

char*	getenv();			/*Define Lattice getenv() function*/


int _showds()				/*Obtain DS register value*/
{
  struct SREGS sfile;			/*Temporary segment register file*/

  segread(&sfile);			/*Initialize segment register file*/

  return(sfile.ds);			/*Return the DS value*/
}


bool envbool(env,val)			/*Get environment boolean*/
char*	env;
bool	val;
{
  char*	envval = getenv(env);		/*Get environment string (if defined)*/

  if (*envval) val = *envval-'0';	/*If defined, get value (0 or 1)*/

  return(val);				/*Return boolean value of*/
}


mchar envchar(env,val)			/*Get environment character*/
char*	env;
mchar	val;
{
  char*	envval = getenv(env);		/*Get environment string (if defined)*/

  if (*envval) val = *envval;		/*If defined, get 1st character*/

  return(val);				/*Return character value*/
}


ci()					/*Quick and dirty 'console input'*/
{
  return(bdos(7) & 0xFF);
}


movmem(x,y,s)				/*Move memory*/
char*	x;				/*Provided due to bug in Lattice C*/
char*	y;				/*runtime (bug is reg ES not set in*/
int	s;				/*movmem(). It is innocuous on TI Pro*/
{					/*and IBM PC, but not in ITT XTRA*/
  while (s-- > 0) *y++ = *x++;
}


sysint(interrupt,r_in,r_out)		/*Interface sysint() to int86x()*/
int	interrupt;
SYSREG* r_in;
SYSREG* r_out;
{
  int	result;

  struct XREG  rfile;
  struct SREGS sfile;

  segread(&sfile);

  rfile.ax = r_in->ax;
  rfile.bx = r_in->bx;
  rfile.cx = r_in->cx;
  rfile.dx = r_in->dx;
  rfile.si = r_in->si;
  rfile.di = r_in->di;

  sfile.ds = r_in->ds;
  sfile.es = r_in->es;

  result = int86x(interrupt,&rfile,&rfile,&sfile);

  r_out->ax = rfile.ax;
  r_out->bx = rfile.bx;
  r_out->cx = rfile.cx;
  r_out->dx = rfile.dx;
  r_out->si = rfile.si;
  r_out->di = rfile.di;

  r_out->ds = sfile.ds;
  r_out->es = sfile.es;

  return(result);
}

#else
/******************************************************************************/
/*									      */
/*			 Compiler-specific code for			      */
/*									      */
/*		        Unknown compiler (you decide)			      */
/*									      */
/******************************************************************************/

bool envbool(env,val)			/*Get environment boolean*/
char*	env;
bool	val;
{
  return(val);
}

mchar envchar(env,val)			/*Get environment character*/
char*	env;
mchar	val;
{
  return(val);
}

/******************************************************************************/
#endif
#endif
#endif
/******************************************************************************/
/*									      */
/*			      Heap sort package   			      */
/*									      */
/* This heapsort package is used because:				      */
/*									      */
/*	1) It's fast 							      */
/*	2) It's non-recursive (no problems with stack overflow)		      */
/*	3) qsort() function available in CI86 and DeSmet C and used for       */
/*	   sorting in the previous versions of ls is not available in         */
/*	   Lattice C, which was used for the present version of ls.	      */
/******************************************************************************/
#define ptr(ndx) (obuf + (ndx)*size)	/*Computes indices into sort buffer*/

hsort(obuf,n,size,comp)			/*Top level of heap sort*/
char*	obuf;
int	n;
int	size;
int	(*comp)();
{
  int	indx;

  for (indx = n/2; indx > 0; indx--) {	/*Scan sort buffer and...*/
    hsort_adjust(obuf,indx,n,size,comp);/*convert it to a heap*/
  }					/*End of convert-to-heap*/

  for (indx = n-1; indx > 0; indx--) {	/*Sort heap by working top down...*/
    swapmem(ptr(0),ptr(indx),size);     /*Swap zero'th & current entry in scan*/
    hsort_adjust(obuf,1,indx,size,comp);/*Recreate the new, smaller heap*/	
  }
}

hsort_adjust(obuf,i,n,size,comp)	/*Heap conversion function*/
char*	obuf;
int	i;
int	n;
int	size;
int	(*comp)();
{
  int	j;
  char	temp[512];

  obuf -= size;				/*Adjust for zero-based indexing*/

  j = 2 * i;

  movmem(ptr(i),&temp,size);

  while (j <= n) {			/*find max of left and right child*/
    if ((j < n) &&
        (*comp)(ptr(j),ptr(j+1)) <= 0) j++;
    if ((*comp)(&temp,ptr(j)) > 0) break;
    movmem(ptr(j),ptr(j/2),size);
    j *= 2;
  }
  movmem(&temp,ptr(j/2),size);
  return;
}

swapmem(x,y,s)				/*Swap memory*/
char*	x;
char*	y;
int	s;
{
  char	t;

  while (s-- > 0) {
    t    = *x;
    *x++ = *y;
    *y++ = t;
  }
}
/******************************************************************************/
